###
###  Copyright (c) 2019 - 2024 Curtis Leach.  All rights reserved.
###
###  Module: Fred::Fish::DBUG::DEBUG

package Fred::Fish::DBUG::DEBUG;

use strict;
use warnings;

use vars qw ( @ISA @EXPORT @EXPORT_OK $VERSION );
use Exporter;

$VERSION   = "2.06";
@ISA       = qw ( Exporter );
@EXPORT    = qw ();
@EXPORT_OK = qw ();

# =============================================================================
# Only POD text appears below this line!
# =============================================================================

=head1 NAME

Fred::Fish::DBUG::DEBUG - A Fred Fish library extension.

=head1 DESCRIPTION

This extension has been removed.  See L<Fred::Fish::DBUG> instead.

=cut

# =============================================================================
#required if module is included w/ require command;
1;
